/*
 * Licensed under the GNU General Public License, version 2.
 *
 * (c) 2005 Craig Duncan Willis <craig-duncan@earthlink.net>
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <ogg/ogg.h>
#include <sys/stat.h>
#include <vorbis/codec.h>
#include <errno.h>
#include <string.h>
#include <limits.h>

#include <locale.h>
#include "i18n.h"

#define RDBUFSZ 4096
#define OV_EWRITE (OV_ENOSEEK-1)

static int writeSerialNumber = 0;    // Isn't serno if 0 or -1, otherwise, is.
static int write_cached_headers = 0;

struct _reportingFlags
{
   unsigned int displayPacket : 1;
   unsigned int displayPage : 1;
   unsigned int displayStream : 1;
   unsigned int debugOutput : 1;
   unsigned int showWarnings : 1;
   unsigned int reportAll : 1;

} reportingFlags;

struct _optionFlags
{
   unsigned int displaySerno : 1;
   unsigned int overwriteOutputFile : 1;
   unsigned int skipWritingHeaders : 1;
   unsigned int writeHeadersOnly : 1;
   unsigned int interactivePageWriteMode : 1;
   unsigned int sampleAccurate : 1;
   unsigned int fetchFailOnHole : 1;
   unsigned int flushAfterFirstTwoPackets : 1;
} optionFlags;

void
initOptions()
{
   reportingFlags.displayPacket = 0;
   reportingFlags.displayPage = 0;
   reportingFlags.displayStream = 0;
   reportingFlags.debugOutput = 0;
   reportingFlags.showWarnings = 0;
   reportingFlags.reportAll = 0;

   optionFlags.displaySerno = 0;
   optionFlags.overwriteOutputFile = 0;
   optionFlags.skipWritingHeaders = 0;
   optionFlags.writeHeadersOnly = 0;
   optionFlags.interactivePageWriteMode = 0;
   optionFlags.sampleAccurate = 1;
   optionFlags.fetchFailOnHole = 0;
   optionFlags.flushAfterFirstTwoPackets = 1;
}


typedef struct
{
   int length;
   unsigned char *data;
} vPacket;


typedef struct _vStream
{
   ogg_stream_state *ogg_stream;
   long serialno;
   ogg_int64_t  initial_granpos;
   ogg_int64_t  cum_samples;    // cumulative number of samples read without
                                // encountering any stream discontinuities
   ogg_int64_t  samplesInPage;
   double       prevStreamSecs; // number of seconds of audio in all previous streams
   int initialGranposSet;
   int  prevW;
   vorbis_info *vi;
// vorbis_dsp_state *vd;
// vorbis_block     *vb;
   vPacket *header[3];  // for saving first 3 vorbis headers
   ogg_packet lastPacket;
} vStream;

typedef struct _oggy
{
   char sig[4];
   ogg_sync_state   *ogg_sync;
   FILE             *fp;
   struct _vStream   *stream;
   ogg_int64_t  page_granulepos;
   int          page_packets;
   long         page_pageno;
   int          newStream;
   int          loadingStreamHeaders;
   struct _oggy *input_oggy;   // used by output oggy to find input oggy headers
} oggy;


static int _cache_ogg_headers(oggy *);



//--------------------------------------------------
// USAGE
//--------------------------------------------------
void
usage( int shouldExit )
{
   printf( _("Vorbis EXtract -- Extract a section from an ogg stream.\n\n") );
   printf( _("Usage: vex <infile.ogg> [ <begin-time> ] [ <end-time> ] [ <outfile.ogg> ]\n\n") );
   if ( shouldExit )
      printf( _("\"vex -h\" for options.\n") );
           
   if ( shouldExit )
      exit(1);
}


//--------------------------------------------------
// fullUsage
//--------------------------------------------------
fullUsage()
{
   usage(0);
   printf( _("    <infile> may be '-' in which case input is taken from stdin.\n") );
   printf( _("    <begin-time> defaults to '0', <end-time> defaults to '-'\n") );
   printf( _("    <begin-time> must be specified if <end-time> is (duh).\n") );
   printf( _("    To have program extract and write a section from the input stream,\n") );
   printf( _("    4th arg must be specified, thus making 2nd & 3rd args mandatory.\n") );
   printf("\nDisplay options:\n");
   printf("    -d     Display debug output\n");
   printf("   -pk     Display info on each packet\n");
   printf("   -pg     Display info on each page\n");
   printf("    -s     Display serial # and stream info\n");
   printf("   -st     Display visual info on stream ('.' per page)\n");
   printf("    -w     Show warnings\n");
   printf("\nOther options:\n");
   printf("    -f     Overwrite existing output file\n");
   printf("   -ff     \"fetch_next_packet\" returns OV_HOLE on encountering hole in stream\n");
   printf("   -nf     Don't write only 2 packets in first audio page (per Ogg/Vorbis spec)\n");
   printf("   -ho     Write stream headers _only_\n");
   printf("   -nh     Write stream _without_ header info\n");
   printf("  -sin     Disable rewriting granulepos at start/end of stream for sample-accurate cutting\n");
   printf("  -ipm     Interactive page writing mode\n");
   printf("   -ra     Put display options into effect throughout entire processing of file\n");
   printf("           (not just within the time interval specified).\n");
   printf(" --serial-number <serial number>   Write output stream with specified serial number.\n");
   printf("           Specifying -1 sets to same as input stream.  Otherwise, random serial\n");
   printf("           numbers are generated as needed.\n");
   printf("    -h     This help\n");
   exit(0);
}


fileExists( char *fnm )
{
   struct stat sbuf;
   return !stat( fnm, &sbuf );
}


//--------------------------------------------------
// randomSerialno
//--------------------------------------------------
long
randomSerialno()
{
   static int seeded = 0;
   if ( !seeded )
   {
      srand( time(NULL) );
      seeded = 1;
   }
   return rand();
}

//--------------------------------------------------
// displayPacket
//--------------------------------------------------
void
displayPacket( oggy *os, ogg_packet *op )
{
   printf("packet#: %6lld    len: %5d   last sample: %6lld\n",
          op->packetno, op->bytes, os->stream->cum_samples);
}

//--------------------------------------------------
// showRelVals
//--------------------------------------------------
showRelVals( float a, float b, float c )
{
   printf("%f < %f <= %f secs (%f + %f = %f)\n", a, b, c, b - a, c - b, c - a);
}

//--------------------------------------------------
// queryToContinue
//--------------------------------------------------
queryToContinue( char *str )
{
   int c;
   printf( "%s ", str);
   system("stty raw -echo");
   c = getchar();
   system("stty -raw echo");
   switch ( c )
   {
      case 'y': case ' ': return 1;
      case 'q': return -1;
      default: return 0;
   }
}


void
IPM( oggy *or, oggy *ow )
{
   int rc;
   ogg_page page;
   while ( get_next_page( or, &page ) > 0 && queryToContinue("skip?") )
      ;
   while ( get_next_page( or, &page ) > 0 )
      if ( (rc = queryToContinue("write?")) > 0 )
         _write_actual_page_to_file( &page, ow->fp );
      else if ( rc == 0 )
         printf("skipped write\n");
      else
         break;
}


//--------------------------------------------------
// mkOggPacket
//--------------------------------------------------
ogg_packet *
mkOggPacket( vPacket *vp )
{
   static ogg_packet pkt;
   memset( &pkt, 0, sizeof(ogg_packet) );
   pkt.packet = vp->data;
   pkt.bytes = vp->length;
   pkt.b_o_s = 1;
   return &pkt;
}


//--------------------------------------------------
// copy goes => way (1st arg into 2nd arg)
//--------------------------------------------------
void
_copyPacketData( ogg_packet *p, ogg_packet *sav )
{
   if ( sav->packet == NULL || p->bytes > sav->bytes )
   {
      if ( sav->packet )
         free( sav->packet );
      sav->packet = malloc( p->bytes );
   }
   memcpy( sav->packet, p->packet, p->bytes );
   sav->bytes = p->bytes;
   sav->granulepos = p->granulepos;
}


static vPacket *
_savePacket( ogg_packet *packet )
{
   vPacket *p = malloc( sizeof(vPacket) );

   p->length = packet->bytes;
   p->data = malloc( p->length );
   memcpy( p->data, packet->packet, p->length );

   return p;
}

_freePacket( vPacket **vp )
{
   if ( *vp && (*vp)->data )
      free( (*vp)->data );
   if ( *vp )
   {
      free( *vp );
      *vp = NULL;
   }
}


static long
_packet_blocksize( oggy *os, ogg_packet *op )
{
   // This returns (approximately) the number of samples in the packet.
   //
   //   From vorbis_synthesis_blockin():
   //   codec_setup_info *ci = vi->codec_setup;
   //     if ( vd->granulepos != -1 )
   //        vd->granulepos += ci->blocksizes[vd->lW] / 4 + ci->blocksizes[vd->W] / 4;

   int ret, thisW = vorbis_packet_blocksize( os->stream->vi, op );
   if ( thisW <= 0 )  // Probably returned OV_ENOTAUDIO
      return 0;

   // don't know if this is actually better but if we're going to assume
   // all packets are of same size in order to work backwards from the
   // initial granulepos of the stream to the gp of the 1st packet, then
   // this first blocksize should be more like the rest than the first.
   if ( os->stream->prevW == 0 )
      os->stream->prevW = thisW;

   ret = ( thisW + os->stream->prevW ) / 4;
   os->stream->prevW = thisW;
   return ret;
}

//--------------------------------------------------
// Returns elapsed time to this point in the
// stream(s) in seconds
//--------------------------------------------------
double
curr_time( oggy *os )
{
   vStream *s = os->stream;
   double secs = 0;
   ogg_int64_t relSampleNo;

   if ( s == NULL || s->vi->rate == 0 )
      return s ? s->prevStreamSecs : 0;

   // Both cum_samples (relative to initial_granpos) and initial_granpos
   // will not be properly init'd until last packet of first audio page
   // of a stream is processed.

   if ( !s->initialGranposSet
        ||
        (relSampleNo = s->cum_samples - s->initial_granpos) < 0 )
   {
      relSampleNo = s->cum_samples;
   }
   secs = s->prevStreamSecs + (double)relSampleNo / (double)s->vi->rate;
   return secs;
}


//--------------------------------------------------
// secs_to_timeSpec
//--------------------------------------------------
char *
secs_to_timeSpec( double secs )
{
   static char buf[32];
   if ( secs >= 60.0 )
   {
      int mins = secs / 60.0;
      secs -= mins * 60;
      if ( secs > 59.99499999 )
      {
         mins++;
         secs = 0.0;
      }
      sprintf(buf, "%d:%05.2f", mins, secs);
   }
   else
      sprintf(buf, "%.2f secs", secs);

   return buf;
}


//--------------------------------------------------
// dumpStream
//--------------------------------------------------
dumpStream( oggy *os )
{
   vStream *s = os->stream;
   printf("init granpos: %10lld, cum: %10lld, prevSecs: %10f\n",
          s->initial_granpos, s->cum_samples, s->prevStreamSecs);
}


//--------------------------------------------------
// fetch_fail_on_hole
//
// This isn't used by this program but allows
// "fetch_next_packet" access to stream provided
// by this program to signal any holes encountered
// if caller desires that (instead of what this
// program wants, which is to pretend a new
// stream started but otherwise ignore the hole).
//--------------------------------------------------
void
fetch_fail_on_hole( int arg )
{
   optionFlags.fetchFailOnHole = arg;
}


//--------------------------------------------------
// _newStream
//--------------------------------------------------
vStream *
_newStream( long serno, oggy *os, int newStream )
{
   vStream *s;

   if ( os->stream )
   {
      s = os->stream;
      s->prevStreamSecs = curr_time(os);
      if ( newStream )
      {
         ogg_stream_clear( s->ogg_stream );
         vorbis_info_clear( s->vi );
         os->newStream = 1;
      }
   }
   else
   {
      s = malloc( sizeof(vStream) );
      memset( s, 0, sizeof(vStream) );
      s->ogg_stream = malloc( sizeof(ogg_stream_state) );
      s->vi         = malloc( sizeof(vorbis_info) );
   }

   s->initialGranposSet = 0;
   s->initial_granpos   = 0;
   s->cum_samples       = 0;

   if ( newStream || os->stream == NULL )
   {
      ogg_stream_init( s->ogg_stream, s->serialno = serno );

      // Need vorbis_info struct in order to initialize
      // codebooks (and get # channels and bit rate).
      vorbis_info_init( s->vi );
   }

   if ( reportingFlags.displayStream && os->stream )
      printf("|%s|\n", secs_to_timeSpec( curr_time(os) ));

   if ( optionFlags.displaySerno && newStream )
   {
      int i; for ( i=0; i < 40; i++ ) putchar('*'); putchar('\n');
      printf("          serialno: %ld\n", serno);
      for ( i=0; i < 40; i++ ) putchar('*'); putchar('\n');
   }

   return s;
}


//--------------------------------------------------
// _isNewStream
//--------------------------------------------------
_isNewStream( long serno, oggy *os )
{
   if ( os->page_pageno == 0 && os->page_packets == 1 && os->page_granulepos == 0 )
      return 1;
   if ( os->stream == NULL )
      return 1;
   if ( serno != os->stream->ogg_stream->serialno )
      return 1;
   return 0;
}


//----------------------------------------------------------
// reportStreamDiscontinuity - 
//
// 2 of 3 calls to this function normally don't do anything
// because the discontinuity has already been noted.
// Note "fetch_fail_on_hole" function, though.
//----------------------------------------------------------
int
reportStreamDiscontinuity( oggy *os, long prevPageno )
{
   if ( prevPageno > 0 && reportingFlags.showWarnings )
   {
      int eos = ogg_stream_eos( os->stream->ogg_stream );

      printf("%s encountered at %s", eos ? "eos" : "stream discontinuity",
             secs_to_timeSpec( curr_time(os) ));

      if ( prevPageno > 0 )
         printf(": page %d => %d\n", prevPageno, os->page_pageno);
      else
         putchar('\n');
   }
   return optionFlags.fetchFailOnHole ? OV_HOLE : 1;
}


//--------------------------------------------------
// _write_actual_page_to_file
//--------------------------------------------------
int
_write_actual_page_to_file( ogg_page *og, FILE *fp )
{
   if ( fwrite(og->header, 1, og->header_len, fp) == og->header_len
        &&
        fwrite(og->body, 1, og->body_len, fp) == og->body_len )
      return 0;
   return -1;
}


//--------------------------------------------------
// _write_pages_to_file
// Returns 0 for success, or -1 on failure. */
//--------------------------------------------------
static int
_write_pages_to_file( ogg_stream_state *stream, FILE *fp, int flush )
{
   ogg_page page;
   static long lastPageno = 0;

   int (*f)() = ( flush ? ogg_stream_flush : ogg_stream_pageout );

   while ( (*f)(stream, &page) )
      if ( _write_actual_page_to_file( &page, fp ) < 0 )
         return -1;

   return 0;
}


//--------------------------------------------------
// _submit_headers_to_stream
//--------------------------------------------------
static void
_submit_headers_to_stream( vPacket **hdrp, ogg_stream_state *stream )
{
   int i;
   for ( i=0; i < 3; i++ )
   {
      ogg_packet p;
      p.bytes = hdrp[i]->length;
      p.packet = hdrp[i]->data;
      // p.b_o_s doesn't need to be set.  Handled automatically by underlying stream functions
      p.e_o_s = 0;
      p.granulepos = 0;

      ogg_stream_packetin( stream, &p );
   }
}


//--------------------------------------------------
// get_next_page
//--------------------------------------------------
int
get_next_page( oggy *os, ogg_page *og )
{
   int i, rc, prevPageno;
   ogg_int64_t prevGranpos;

   for ( i=0; i < 20; i++ )  // if we haven't synced after this many reads...
   {
      if ( (rc = ogg_sync_pageout( os->ogg_sync, og )) > 0 )
         break;

      if ( rc == 0 )   // Need more data from ogg stream
      {
         unsigned char *buf = ogg_sync_buffer( os->ogg_sync, RDBUFSZ );
         int bytes = fread( buf, 1, RDBUFSZ, os->fp );
         if ( bytes > 0 )
         {
            ogg_sync_wrote( os->ogg_sync, bytes );
         }
         else if ( ferror(os->fp) )
         {
            printf( _("Error reading stream: %s\n"), strerror(errno) );
         }
         else if ( feof(os->fp) )
         {
            if ( reportingFlags.showWarnings )
            {
               if ( !ogg_stream_eos( os->stream->ogg_stream ) )
                  printf( _("\nOgg stream ended abruptly (no <eos> received)\n") );
               else
                  printf( _("\nEnd of ogg stream reached\n") );
            }
            break;
         }
      }
      else  // rc < 0
      {
         if ( os->stream == NULL || os->stream->serialno == -1 )
         {
            // We haven't got a serial number yet which means we've failed
            // to get the first header packet at the start of the stream.
            printf( _("Input stream not ogg.\n") );
            return OV_ENOTVORBIS;
         }
         else
            printf( _("Lost page sync, input stream corrupted\n") );
      }
   }

   if ( rc != 1 )
      return feof(os->fp) ? OV_EOF : OV_EREAD;

   // Got page.
   prevPageno  = os->page_pageno;
   prevGranpos = os->page_granulepos;
   os->page_pageno     = ogg_page_pageno( og );
   os->page_packets    = ogg_page_packets( og );
   os->page_granulepos = ogg_page_granulepos( og );

   if ( reportingFlags.debugOutput && ogg_page_eos(og) )
      printf("<pg eos>\n");

   long serno = ogg_page_serialno(og);
   int newStream = _isNewStream( serno, os );

   if ( reportingFlags.displayPage )
      printf("%s*** page %3d:  len: %d/%d   %2d packets (gp: %lld)\n",
             reportingFlags.displayPacket ? "\n" : "",
             os->page_pageno, og->header_len, og->body_len,
             os->page_packets, os->page_granulepos);
   else if ( reportingFlags.displayStream )
      if ( newStream ) fputs("\n^", stdout);
      else putchar('.');

   if ( ++prevPageno != os->page_pageno )
   {
      reportStreamDiscontinuity(os, prevPageno-1);   // hole in stream or new stream
      os->stream = _newStream( serno, os, newStream );
   }

   return 1;
}


//--------------------------------------------------
// Returns:
//    1   when packet has been read
//   <0   OV_EOF or error
//        OV_HOLE if "fetch_fail_on_hole(1)" called
//--------------------------------------------------
int
fetch_next_packet( oggy *os, ogg_packet *op )
{
   int rc = 0;
   long bs;
   static ogg_page page;

   if ( os == NULL || memcmp( &os->sig, "OGGR", 4 ) != 0 )
      return OV_EINVAL;

   if ( os->stream )
   {
      if ( ogg_stream_eos( os->stream->ogg_stream ) && reportingFlags.displayStream )
         putchar('*');

      while ( (rc = ogg_stream_packetout( os->stream->ogg_stream, op )) < 0 )
         if ( reportStreamDiscontinuity(os, -1) < 0 )
            return OV_HOLE;

      if ( rc > 0 )
         goto gotPacket;
   }

   if ( (rc = get_next_page( os, &page )) < 0 )
      return rc;

   if ( ogg_stream_pagein( os->stream->ogg_stream, &page ) < 0 )
   {
      // above succeeds when there's a hole (ogg_stream_packetout returns -1)
      if ( ogg_page_version(&page) != 0 )
      {
         printf( _("Ogg version %d not supported\n"), ogg_page_version(&page) );
         return OV_EVERSION;
      }  
      printf( "Internal error\n" );
      return OV_EFAULT;
   }

   if ( os->newStream && !os->loadingStreamHeaders )
   {
      if ( _cache_ogg_headers( os ) < 0 )
      {
         if ( os->stream->lastPacket.packet == NULL )
            return OV_EBADHEADER;
         memcpy( op, &os->stream->lastPacket, sizeof(ogg_packet) );
         goto gotPacket;
      }
      else
      {
         write_cached_headers = 1;
         return fetch_next_packet(os, op);
      }
   }

   while ( (rc = ogg_stream_packetout( os->stream->ogg_stream, op )) < 0 )
      if ( reportStreamDiscontinuity(os, -1) < 0 )
         return OV_HOLE;

   if ( rc == 0 )  // No packets available
      return ( ogg_stream_eos(os->stream->ogg_stream) ? OV_EOF : OV_EREAD );

gotPacket:
   if ( !os->loadingStreamHeaders )   // Don't do next stuff for header packets
   {
      bs = _packet_blocksize( os, op );
      if ( bs == 0 && reportingFlags.showWarnings )
         printf("blocksize of 0\n");

      // Increment cumulative #samples (since initial granulepos value)
      // by (approximate?) number of samples in this packet.
      os->stream->cum_samples += bs;

      if ( !os->stream->initialGranposSet )  // 1st (audio) page of stream
         // sum blocksize of all packets in first audio page in order to subtract
         // from granulepos of page to get granulepos of very first data packet.
         os->stream->samplesInPage += bs;

      if ( op->granulepos > 0 )    // Last packet in page (has the actual GP)
      {
         if ( os->stream->cum_samples != op->granulepos )
         {
            if ( reportingFlags.showWarnings )
            {
               if ( !os->stream->initialGranposSet )      // Last packet of first page
                  printf("Initial GP short (page GP: %lld, samples in page: %lld)\n",
                         op->granulepos, os->stream->samplesInPage);
               else if ( ogg_stream_eos( os->stream->ogg_stream ) ) // Last packet of last page
                  printf("Last GP short (page GP: %lld, last sample: %lld)\n",
                         op->granulepos, os->stream->cum_samples);
               else // Mid-stream.  Granulepos jumped discontinuously (while pageno didn't).
                  printf("Granulepos discontinuity: Calculated (%lld) != actual (%lld)"
                         " for last packet in page %ld\n",
                         os->stream->cum_samples, op->granulepos, os->page_pageno);
            }

            os->stream->cum_samples = op->granulepos;
         }

         if ( !os->stream->initialGranposSet )  // 1st (audio) page of stream
         {
            // Last packet of first page
            os->stream->initial_granpos = op->granulepos - os->stream->samplesInPage;
            if ( os->stream->initial_granpos < 0 )
               os->stream->initial_granpos = 0;
            os->stream->initialGranposSet = 1;
            os->stream->samplesInPage = 0;
         }
         if ( reportingFlags.debugOutput && op->e_o_s )
            printf("<pk eos>\n");
      }
      else
         op->granulepos = os->stream->cum_samples; // _Every_ packet gets granulepos
   }

   if ( reportingFlags.displayPacket )
      displayPacket(os, op);
   return 1;
}


//--------------------------------------------------
// _cache_ogg_headers
//--------------------------------------------------
static int
_cache_ogg_headers( oggy *os )
{
   vorbis_comment vc;
   ogg_page page;
   ogg_packet packet;
   int i, rc, bytes;
   unsigned char *buffer;

   os->newStream = 0;
   os->loadingStreamHeaders = 1;
   // Need to process comments, even if there are no user
   // comments, in order to create 2nd (comment) output header.
   vorbis_comment_init( &vc );

   for ( i = 0; i < 3; i++ )  // Three headers to process: Vorbis ID, Comments & Codebooks
   {
      if ( fetch_next_packet( os, &packet ) < 0 )
      {
         char *msg = ( i == 0 ? _("Error in input stream") : _("Secondary header corrupt") );
         printf( "%s.\n", msg );
         i = 1;  // So we don't assume that we read first packet ok but synthesis failed
         break;
      }

      if ( vorbis_synthesis_headerin( os->stream->vi, &vc, &packet ) < 0 )
      {
         printf( _("Error in %s at %s\n"), i == 0 ?
                 _("primary header (not vorbis?)") : _("secondary header"),
                 secs_to_timeSpec( curr_time(os) ) );
         break;
      }

      _freePacket( &os->stream->header[i] );
      os->stream->header[i] = _savePacket( &packet );
   }

   if ( i == 0 && os->stream->header[i] )
   {
      // Maybe we haven't really encountered a new stream.
      // "Reload" (and continue using) original headers and "return"
      // the packet just read.  This may cause us to continue
      // processing a bad stream so warn about this potential.
      int j;
      for ( j = 0; j < 3; j++ )
         vorbis_synthesis_headerin( os->stream->vi, &vc, mkOggPacket(os->stream->header[j]) );
      memcpy( &os->stream->lastPacket, &packet, sizeof(ogg_packet) );

      printf("Continuing processing using original headers.\n"
             "  (Warning: Stream may be too corrupt to give meaningful results.)\n");
   }

   os->loadingStreamHeaders = 0;
   vorbis_comment_clear( &vc );

   if ( i < 3 )
   {
      // If any header was bad, delete them all (we might have failed to
      // load new headers but don't want to delete previous, valid, set).
      if ( os->stream && os->stream->header[2] == NULL )
         for ( i = 0; i < 2; i++ )
            _freePacket( &os->stream->header[i] );
      return OV_EBADHEADER;
   }

   if ( optionFlags.displaySerno )
   {
      long rate = os->stream->vi->rate;
      long channels = os->stream->vi->channels;
      long upper = os->stream->vi->bitrate_upper;
      long nominal = os->stream->vi->bitrate_nominal;
      long lower = os->stream->vi->bitrate_lower;
      long window = os->stream->vi->bitrate_window;
      printf("<rate: %d, ch#: %d", rate, channels);
      if ( upper > 0 )
         printf(", upper: %d", upper);
      if ( nominal > 0 )
         printf(", nominal: %d", nominal);
      if ( lower > 0 )
         printf(", lower: %d", lower);
      if ( window > 0 )
         printf(", window: %d", window);
      fputs(">\n", stdout);
   }  

   return 0;
}


//--------------------------------------------------
// write_packet
//
// This is a mess in order that the last packet can be marked as
// end-of-stream.  This requires saving each packet passed in and
// writing the packet _previously_ passed.  The need to copy the packet
// makes this a good place to rewrite the packet granulepos, though.
//
// RETURNS 0 on success.
//--------------------------------------------------
int
write_packet( ogg_packet *op, oggy *os )
{
   static ogg_packet* packetToWrite = NULL;
   int dontDoWrite = 0;

   if ( os == NULL || memcmp( &os->sig, "OGGW", 4 ) != 0 )
      return OV_EINVAL;

   if ( os->stream == NULL || write_cached_headers )
   {
      write_cached_headers = 0;

      if ( os->stream )
      {
         ogg_flush(os);
         ogg_stream_clear( os->stream->ogg_stream );
         os->page_packets = 0;
      }
      else
      {
         os->stream = malloc( sizeof(vStream) );
         memset( os->stream, 0, sizeof(vStream) );
         os->stream->ogg_stream = malloc( sizeof(ogg_stream_state) );
      }
   
      if ( writeSerialNumber == 0 )
         os->stream->serialno = randomSerialno();
      else if ( writeSerialNumber == -1 )
         os->stream->serialno = os->input_oggy->stream->ogg_stream->serialno;

      ogg_stream_init( os->stream->ogg_stream,
                       writeSerialNumber > 0 ? writeSerialNumber : os->stream->serialno );
   
      if ( !optionFlags.skipWritingHeaders )
      {
         _submit_headers_to_stream( os->input_oggy->stream->header, os->stream->ogg_stream );

         if ( _write_pages_to_file( os->stream->ogg_stream, os->fp, 1 ) == -1
              ||
              optionFlags.writeHeadersOnly )
         {
            ogg_close( os );
            return optionFlags.writeHeadersOnly ? OV_EOF : OV_EWRITE;
         }
      }
   }

   if ( packetToWrite == NULL )
   {
      packetToWrite = malloc( sizeof(ogg_packet) );
      memset( packetToWrite, 0, sizeof(ogg_packet) );
   }

   if ( os->stream->lastPacket.packet == NULL )
   {
      dontDoWrite = 1;
      os->stream->initial_granpos = op->granulepos -
         (os->page_granulepos >= 0 ? os->page_granulepos : _packet_blocksize(os->input_oggy, op));
   }
   else
      _copyPacketData( &os->stream->lastPacket, packetToWrite );

   _copyPacketData( op, &os->stream->lastPacket );
   os->stream->lastPacket.granulepos -= os->stream->initial_granpos;

   if ( dontDoWrite )
      return 0;

   packetToWrite->e_o_s = 0;
   os->page_packets++;
   ogg_stream_packetin( os->stream->ogg_stream, packetToWrite );
   // From "Ogg Vorbis I format specification:
   //           Embedding Vorbis into an Ogg Stream":
   // In both of these cases in which the initial audio PCM starting
   // offset is nonzero, the second finished audio packet must flush
   // the page on which it appears and the third packet being a fresh
   // page.  This allows the decoder to always be able to perform PCM
   // position adjustments before needing to return any PCM data from
   // synthesis, resulting in correct positioning information without
   // any additional seeking logic.
   {
      int flush = 0;
      if ( os->page_packets == 2 && optionFlags.flushAfterFirstTwoPackets )
         flush = 1;
      return _write_pages_to_file( os->stream->ogg_stream, os->fp, flush );
   }
}


//--------------------------------------------------
// ogg_flush
//--------------------------------------------------
int
ogg_flush( oggy *os )
{
   if ( os == NULL || memcmp( &os->sig, "OGGW", 4 ) != 0 )
      return OV_EINVAL;

   if ( os->stream->lastPacket.packet == NULL )
      return 0;

   os->stream->lastPacket.e_o_s = 1;
   ogg_stream_packetin( os->stream->ogg_stream, &os->stream->lastPacket );
   if ( _write_pages_to_file( os->stream->ogg_stream, os->fp, 1 ) < 0 )
      printf("Error flushing to output file\n");
   free( os->stream->lastPacket.packet );
   os->stream->lastPacket.packet = NULL;
}

//--------------------------------------------------
// ogg_close
//--------------------------------------------------
int
ogg_close( oggy *os )
{
   if ( os == NULL || memcmp( &os->sig, "OGG", 3 ) != 0
        ||
        (os->sig[3] != 'R' && os->sig[3] != 'W') )
      return -1;

   if ( os->stream && os->stream->ogg_stream )
   {
      if ( os->sig[3] == 'W' )
         ogg_flush( os );
      ogg_stream_destroy( os->stream->ogg_stream );
   }
   if ( os->fp )
      fclose( os->fp );

   if ( os->stream && os->stream->vi )
   {
      vorbis_info_clear( os->stream->vi );
      free( os->stream->vi );
   }

   if ( os->ogg_sync )
      ogg_sync_destroy( os->ogg_sync );

//  Use: vorbis_dsp_clear(vd) + vorbis_block_clear(vb), then free(vd), free(vb)

   free(os);
   return 0;
}


//--------------------------------------------------
// ogg_open_writable
//--------------------------------------------------
oggy *
ogg_open_writable( char *fnm, oggy *or )
{
   FILE *fp;

   if ( or == NULL || memcmp( &or->sig, "OGGR", 4 ) != 0 )
      return NULL;

   if ( strcmp( fnm, "-" ) != 0 )
   {
      fp = fopen( fnm, "wb" );
      if ( fp == NULL )
      {
         printf( _("Failed to open %s for reading: %s\n"), fnm, strerror(errno) );
         return NULL;
      }
   }
   else
      fp = stderr;

   oggy *os = malloc( sizeof(oggy) );
   memset( os, 0, sizeof(oggy) );
   memcpy( &os->sig, "OGGW", 4 );

   os->fp = fp;
   os->input_oggy = or;

   return os;
}


//--------------------------------------------------
// ogg_open
//--------------------------------------------------
oggy *
ogg_open( char *fnm, oggy *or )
{
   FILE *fp;

   if ( or != NULL )
      return ogg_open_writable( fnm, or );

   if ( strcmp( fnm, "-" ) != 0 )
   {
      fp = fopen( fnm, "rb" );
      if ( fp == NULL )
      {
         printf( _("Failed to open %s for reading: %s\n"), fnm, strerror(errno) );
         return NULL;
      }
   }
   else
      fp = stdin;

   oggy *os = malloc( sizeof(oggy) );
   memset( os, 0, sizeof(oggy) );
   memcpy( &os->sig, "OGGR", 4 );

   os->fp = fp;

   os->ogg_sync   = malloc( sizeof(ogg_sync_state) );
   ogg_sync_init( os->ogg_sync );

   if ( _cache_ogg_headers( os ) < 0 )
   {
      ogg_close( os );
      return NULL;
   }

// os->vd = malloc( sizeof(vorbis_dsp_state) );
// vorbis_synthesis_init( os->vd, os->stream->vi );

// os->vb = malloc( sizeof(vorbis_block) );
// vorbis_block_init( os->vd, os->vb );

   return os;
}


//--------------------------------------------------
// timespec_to_secs
//--------------------------------------------------
double
timespec_to_secs( char *timeSpec )
{
   // Format of timeSpec is "<m>:<s>.<f>" (colon and dot optional) where:
   // <m> = minutes : any number, including 0 (or nothing at all before colon)
   // <s> = seconds : '00' => '59' if following a colon.  Anything, otherwise.
   // <f> = fractions of a second : any number or omitted (including the '.')
   // Only seconds part is required.
   int len, colon = 0, mins = 0, secs = 0;
   char *spt, *fpt, fracStr[12], buf[64];
   double frac = 0;

   if ( strcmp( timeSpec, "0" ) == 0 )
      return 0;

   if ( strcmp( timeSpec, "-" ) == 0 )
      return -1;

   // Check that all characters are digits (except possibly one ':' and one '.')
   strcpy( buf, timeSpec );
   spt = strchr( buf, ':' );
   if ( spt )
      *spt = '0';
   fpt = strchr( spt ? spt+1 : buf, '.' );
   if ( fpt )
      *fpt = '0';
   for ( spt=buf; *spt; spt++ )
      if ( !isdigit(*spt) )
         goto bad;

   spt = strchr( timeSpec, ':' );
   if ( spt )
   {
      mins = atoi( timeSpec );
      colon = 1;
      spt++;
   }
   else
      spt = timeSpec;

   fpt = strchr( spt, '.' );
   len = fpt ? fpt - spt : strlen(spt);

   if ( colon && len != 2 )
      goto bad;
   secs = atoi( spt );

   if ( fpt )
   {
      strcpy( fracStr, fpt );
      frac = atof( fracStr );
   }

   return mins * 60 + secs + frac;

 bad:
      printf( _("Invalid time spec: `%s'  Time spec is of form:\n"), timeSpec );
      printf( _("\t[ + ] <m:ss> <m:ss.d> <s.d> <s> <.d>  ('0' is bof, '-' is eof)\n") );
      exit(1);
}


//--------------------------------------------------
// process_args
//--------------------------------------------------
int
process_args( int argc, char **argv )
{
#define CASE(c) else if ( strcmp(pt, c) == 0 )
   char *pt;
   int i, j, k, reduceCount;

   // getopt was too inflexible
   for ( i=1; i < argc; i++ )
   {
      reduceCount = 1;
      pt = argv[i];

      if ( *pt++ != '-' || *pt == '\0' )   // latter is '-' time spec
         continue;

      CASE("d")
      {
         reportingFlags.debugOutput = 1;
         reportingFlags.showWarnings = 1;
         optionFlags.displaySerno = 1;
         reportingFlags.reportAll = 1;
      }
      CASE("f")
         optionFlags.overwriteOutputFile = 1;

      CASE("ff")
         optionFlags.fetchFailOnHole = 1;

      CASE("ho")   // write stream headers _only_
         optionFlags.writeHeadersOnly = 1;

      CASE("h")
         fullUsage();
      CASE("?")
         fullUsage();
      CASE("-help")
         fullUsage();

      CASE("s")
         optionFlags.displaySerno = 1 - optionFlags.displaySerno;

      CASE("ipm")  // Interactive Page Writing Mode
         optionFlags.interactivePageWriteMode = reportingFlags.displayPage = 1;

      CASE("pk")   // Display info on each packet
      {
         reportingFlags.displayPacket = 1;
         reportingFlags.displayPage = 1;
      }

      CASE("pg")   // Display info on each page
         reportingFlags.displayPage = 1;

      CASE("nf")   // Don't flush 1st audio page after two packets (per Ogg/Vorbis spec)
         optionFlags.flushAfterFirstTwoPackets = 0;

      CASE("nh")   // write stream with no header
         optionFlags.skipWritingHeaders = 1;

      CASE("st")   // Display "dots" info for stream
         reportingFlags.displayStream = 1;

      CASE("ra")
         // observe display options for entire file
         // (not just within time interval specified).
         reportingFlags.reportAll = 1;

      CASE("sin")             // disable rewriting of granulepos for
         optionFlags.sampleAccurate = 0;   // sample-accurate (more or less) cutting

      CASE("-serial-number")
      {
         // Use following arg as serial number in stream to be written.
         // Actual numeric must be next arg (can't be trailing part of this arg).
         if ( i+1 == argc || fileExists( argv[i+1] ) )
         {
            // serial number wasn't specified
            printf("Serial number must be next arg after \"-s\" arg\n");
            exit(1);
         }
         writeSerialNumber = atoi( argv[i+1] );
         reduceCount = 2;
      }
      
      CASE("w")
         reportingFlags.showWarnings = 1 - reportingFlags.showWarnings;

      else
      {
         printf("Unknown option: %s\n", --pt);
         exit(1);
      }

      // get rid of arg just processed
      for ( k = 0; k < reduceCount; k++ )
      {
         for ( j=i+1; j < argc; j++ )
            argv[j-1] = argv[j];
         argc--;
      }
      i--;
   }

   return argc;
}


//--------------------------------------------------
// MAIN
//--------------------------------------------------
int
main( int argc, char *argv[] )
{
   struct _reportingFlags savOpts;
   char fnmBuf[PATH_MAX];
   char *outputFilename = NULL;
   char *beginTimeSpec, *endTimeSpec;
   int eof = 0;
   ogg_int64_t lastPacketno, prevGP = 0;
   double targetBeginSecs, targetEndSecs, prevSecs, actualStartSecs, actualEndSecs, tot;
   ogg_packet packet;
   oggy *or, *ow;

   setbuf( stdout, 0 );

   initOptions();
   argc = process_args( argc, argv );
   memcpy( &savOpts, &reportingFlags, sizeof(reportingFlags) );
   if ( !reportingFlags.reportAll )
      memset( &reportingFlags, 0, sizeof(reportingFlags) );

   if ( argc < 2 )
      usage(1);

   // Must have existing ogg file as 1st arg.
   // 2nd and 3rd options are optional but must be timeSpec's.
   // To specify an output file, it must be the 4th arg
   // (which makes 2nd & 3rd args mandatory).
   beginTimeSpec =  argc > 2 ? argv[2] : "0";
   endTimeSpec   =  argc > 3 ? argv[3] : "-";

   targetBeginSecs = timespec_to_secs( beginTimeSpec );
   targetEndSecs   = timespec_to_secs( endTimeSpec[0] == '+' ? endTimeSpec+1 : endTimeSpec );
   if ( endTimeSpec[0] == '+' )
      targetEndSecs += targetBeginSecs;

   if ( targetEndSecs >= 0 && targetBeginSecs > targetEndSecs )
   {
      printf( _("Beginning time after ending time\n") );
      exit(1);
   }

   or = ogg_open( argv[1], NULL );
   if ( or == NULL )
      return 1;

   if ( argc > 4 )
      outputFilename = argv[4];

   if ( outputFilename )
   {
      if ( strcmp( outputFilename + strlen(outputFilename) - 4, ".ogg" ) != 0 )
      {
         sprintf( fnmBuf, "%s.ogg", outputFilename );
         outputFilename = fnmBuf;
         if ( reportingFlags.showWarnings )
            printf(_("Adding \".ogg\" extension to output file (%s)\n"), outputFilename);
      }
      if ( fileExists(outputFilename) && !optionFlags.overwriteOutputFile )
      {
         printf(_("Output file \"%s\" exists.  Use -f to overwrite.\n"), outputFilename);
         return 1;
      }
      ow = ogg_open( outputFilename, or );
      if ( ow == NULL )
      {
         ogg_close(or);
         return 1;
      }
   }
   
   prevSecs = -1;
   actualStartSecs = 0;
   packet.granulepos = 0;

   while ( targetBeginSecs > actualStartSecs )
   {
      prevGP = packet.granulepos;
      prevSecs = actualStartSecs;

      if ( fetch_next_packet( or, &packet ) < 0 )
      {
         printf("Start position not found\n");
         goto finish;
      }
      actualStartSecs = curr_time(or);
   }
   if ( prevSecs == -1 )
      if ( fetch_next_packet( or, &packet ) < 0 )
         printf("Start position not found\n");

   if ( savOpts.displayPacket && !reportingFlags.displayPacket )
      displayPacket(or, &packet);

   if ( !savOpts.reportAll )
      memcpy( &reportingFlags, &savOpts, sizeof(reportingFlags) );

   if ( reportingFlags.debugOutput )
      showRelVals( prevSecs, targetBeginSecs, actualStartSecs );

   if ( optionFlags.sampleAccurate && prevGP > 0 )
   {
      ogg_int64_t startingSample = (targetBeginSecs - prevSecs) / (actualStartSecs - prevSecs) *
         (packet.granulepos - prevGP) + prevGP;

      // borrowing `page_granulepos' to hold number of samples
      // in first audio packet (excluding those before time 0).
      if ( ow )
      {
         ow->page_granulepos = packet.granulepos - startingSample;
         if ( reportingFlags.debugOutput )
            printf("Starting (zero-based) granulepos: %lld\n", ow->page_granulepos);
      }

      if ( reportingFlags.debugOutput )
         printf("GPs: %lld < %lld < %lld\n", prevGP, startingSample, packet.granulepos);
   }
   else if ( ow )
      ow->page_granulepos = -1;  // signal to write_packet not to use this

   if ( optionFlags.interactivePageWriteMode )
   {
      IPM(or, ow);
      goto finish;
   }

   if ( reportingFlags.debugOutput )
      printf("First packet #: %lld\n", packet.packetno);

   actualEndSecs = curr_time(or);
   lastPacketno = packet.packetno;
   prevGP = packet.granulepos;

   while ( targetEndSecs == -1 || actualEndSecs < targetEndSecs )
   {
      if ( outputFilename && write_packet( &packet, ow ) != 0 )
      {
         printf("Write error\n");
         goto finish;
      }

      prevSecs = actualEndSecs;
      prevGP = packet.granulepos;

      if ( fetch_next_packet( or, &packet ) < 0 )
      {
         eof = 1;
         break;
      }
      
      lastPacketno = packet.packetno;
      actualEndSecs = curr_time(or);
   }
   // If the ending sample is in this packet then we write it all
   if ( !eof && outputFilename && write_packet( &packet, ow ) != 0 )
   {
      printf("Write error\n");
      goto finish;
   }

   if ( reportingFlags.debugOutput )
   {
      printf("Last packet #: %lld\n", lastPacketno);
      if ( eof )
         printf("%f secs\n", actualEndSecs);
      else
         showRelVals( prevSecs, targetEndSecs, actualEndSecs );
   }

   /* Generally, the next to last packet's granulepos is before the cut
    * point and the last packet's GP is after the cut point.  So we need
    * to write a short GP in the last packet.  Knowing the end position
    * of the next to last and last packets and the cut point between the
    * two (in seconds), endingSample (between prevGP and latest GP in
    * the same proportion), can be calculated.
    */
   if ( !eof && optionFlags.sampleAccurate )
   {
      ogg_int64_t endingSample = (targetEndSecs - prevSecs) / (actualEndSecs - prevSecs) *
         (packet.granulepos - prevGP) + prevGP;

      or->stream->cum_samples -= (packet.granulepos - endingSample);

      if ( ow )
      {
         ow->stream->lastPacket.granulepos = endingSample - ow->stream->initial_granpos;
         if ( reportingFlags.debugOutput )
            printf("Ending granulepos: %lld\n", ow->stream->lastPacket.granulepos);
      }
      if ( reportingFlags.debugOutput )
         printf("GPs: %lld < %lld < %lld\n", prevGP, endingSample, packet.granulepos);
   }
   
finish:
   // In case we jumped here
   if ( !savOpts.reportAll )
      memcpy( &reportingFlags, &savOpts, sizeof(reportingFlags) );

   if ( reportingFlags.debugOutput )
      dumpStream(or);

   tot = curr_time(or) /* time through to end of last packet read */ - targetBeginSecs;
   printf("total time: %s (%f)\n", secs_to_timeSpec(tot), tot);

   ogg_close(or);
   ogg_close(ow);  // this or 'ogg_flush' makes sure last packet gets written (with eos set)
   return 0;
}

